import weakref


class WeakFunctionReferences:
    def __init__(self):
        self.references = {}

    def add(self, function):
        obj = function.__self__
        fname = function.__name__
        callback_id = self._get_callback_id(function)
        wr = weakref.ref(obj, self._obj_died(callback_id))
        self.references[callback_id] = (wr, fname)
        return callback_id

    def remove(self, function):
        callback_id = self._get_callback_id(function)
        del self.references[callback_id]

    def get(self, callback_id):
        wr, fname = self.references[callback_id]
        obj = wr()
        f = getattr(obj, fname)
        return f

    def get_all(self):
        keys = [key for key in self.references.keys()]
        for key in keys:
            if key in self.references.keys():
                yield self.get(key)

    def _obj_died(self, callback_id):
        def f(wr):
            del self.references[callback_id]

        return f

    @staticmethod
    def _get_callback_id(function):
        obj = function.__self__
        callback_id = hash("%s%s" % (id(obj), function.__name__)) & 0xffffffffffff
        if callback_id < 0:
            callback_id = callback_id * -1
        return callback_id



import jinja2
import weakref
import traceback
import re
import os
import sys
import inspect
import json
import queue
import importlib
import gevent
from .lib import WeakFunctionReferences


class PyHtmlGuiInstance():
    def __init__(self, pyHtmlGui):
        self._pyHtmlGui = pyHtmlGui
        self._appInstance = pyHtmlGui.appInstance
        self._appViewClass = pyHtmlGui.appViewClass
        self._templateLoader = pyHtmlGui._templateLoader
        self._add_file_to_monitor = pyHtmlGui._add_file_to_monitor

        self._websocketInstances = []
        self._children = weakref.WeakSet()
        self._templateEnv = jinja2.Environment(loader=self._templateLoader)
        self._templateCache = {}
        self._call_number = 0
        self._call_javascript = self.call_javascript  # so the call_javascript function in view is available
        self._function_references = WeakFunctionReferences()

        self._templateEnv.globals['_create_py_function_reference'] = self._create_function_reference

        self.view = self._appViewClass(self._appInstance, self)

    @property
    def instances_count(self):
        return len(self._websocketInstances)

    def update(self):
        self.call_javascript("pyhtmlgui.update_element", ["pyHtmlGuiBody", self.view.render()], skip_results=True)

    def set_visible(self, visible):
        if visible is False:
            for child in self._children:
                child.set_visible(False)
        if visible is True:
            self.update()

    # CALL JS FROM PY
    def call_javascript(self, js_function_name, args, skip_results=False):
        self._call_number += 1
        to_delete = self._call_number - 100  # clean old function references, this is needed to results don't stay around if you don't set skip_results=True and don't use the result.
        for websocketInstance in self._websocketInstances:
            if to_delete in websocketInstance.javascript_call_result_objects:
                del websocketInstance.javascript_call_result_objects[to_delete]
            if to_delete in websocketInstance.javascript_call_result_queues:
                del websocketInstance.javascript_call_result_queues[to_delete]

        javascript_call_object = {'call': self._call_number, 'name': js_function_name, 'args': args}

        if skip_results is True:
            javascriptCallResult = None
            javascript_call_object["skip_results"] = True
        else:
            javascriptCallResult = JavascriptCallResult(self._call_number)

        data = json.dumps(javascript_call_object, default=lambda o: None)
        for websocketInstance in self._websocketInstances:
            if skip_results is False:
                websocketInstance.javascript_call_result_objects[self._call_number] = javascriptCallResult
                javascriptCallResult.add_call(websocketInstance)
            websocketInstance.ws.send(data)
        return javascriptCallResult

    # MESSAGES FROM JS TO PY
    def websocket_loop(self, websocketInstance):
        self._websocketInstances.append(websocketInstance)
        while True:
            msg = websocketInstance.ws.receive()
            if msg is not None:
                message = json.loads(msg)
                gevent.spawn(self._websocket_process_message, message, websocketInstance).run()
            else:
                break
        self._websocketInstances.remove(websocketInstance)
        if len(self._websocketInstances) == 0:
            self.set_visible(False)

    def _websocket_process_message(self, message, websocketInstance):
        if 'call' in message:
            function_name = "Function not found"
            args = "  "
            try:
                if message['name'] == "frontend_ready":
                    function_name = "%s.update" % (self.__class__.__name__)
                    return_val = self.update()
                    self._pyHtmlGui._on_frontend_ready(self)

                elif message['name'] == "call_python_function":
                    functioncall_id = message['args'][0]
                    function = self._function_references.get(functioncall_id)
                    function_name = "%s.%s" % (function.__self__.__class__.__name__, function.__name__)
                    return_val = function()

                elif message['name'] == "call_python_function_with_args":
                    functioncall_id = message['args'][0]
                    args = message['args'][1]
                    function = self._function_references.get(functioncall_id)
                    function_name = "%s.%s" % (function.__self__.__class__.__name__, function.__name__)
                    return_val = function(*args)
                else:
                    return_val = None
                    print("unknown python function", message['name'])

            except Exception as e:
                tb = traceback.format_exc()
                msg = " Exception in: %s(%s)\n" % (function_name, ("%s" % args)[1:-1])
                msg += " %s" % tb.replace("\n", "\n  ").strip()
                self.call_javascript("pyhtmlgui.debug_msg", [msg])
                print(msg)
                return_val = None

            if not ("skip_results" in message and message["skip_results"] is True):
                websocketInstance.ws.send(
                    json.dumps({'return': message['call'], 'value': return_val}, default=lambda o: None))

        elif 'return' in message:
            call_id = message['return']
            del message['return']  # remove internal id from result before passing to next level
            if call_id in websocketInstance.javascript_call_result_objects:
                websocketInstance.javascript_call_result_objects[call_id].result_received(websocketInstance, message)
        else:
            print('Invalid message received: ', message)

    def _create_function_reference(self, function):
        if type(function) == jinja2.runtime.Undefined:
            raise Exception("Undefined python method in script: '%s'" % function._undefined_name)
        cbid = self._function_references.add(function)
        return cbid

    def _get_template(self, item, force_reload=False):
        if force_reload is False:
            try:
                return self._templateCache[item.__class__.__name__]
            except:
                pass

        if item.__class__.TEMPLATE_FILE is not None:  # load from file
            file_to_monitor = self._templateEnv.get_template(item.TEMPLATE_FILE).filename
            string_to_render = open(file_to_monitor, "r").read()
        else:  # load from class
            if self._pyHtmlGui.auto_reload is False:
                string_to_render = item.TEMPLATE_STR
                file_to_monitor = None
            else:
                module_name = item.__module__
                if module_name is None or module_name == str.__class__.__module__:
                    module_fullname = item.__class__.__name__  # Avoid reporting __builtin__
                else:
                    module_fullname = module_name + '.' + item.__class__.__name__

                try:
                    file_to_monitor = os.path.abspath(inspect.getfile(item.__class__))
                except:  # in case its in mail, this may be a bug? in inspect
                    file_to_monitor = os.path.abspath(sys.argv[0])

                if module_name == "__main__":
                    name = os.path.splitext(os.path.basename(file_to_monitor))[0]
                    module = __import__(name)
                    importlib.reload(module)  # reload should work on non complex objects in __main__, but not for more
                    for comp in module_fullname.split(".")[1:]:
                        module = getattr(module, comp)
                else:
                    loader = importlib.machinery.SourceFileLoader(module_name, file_to_monitor)
                    spec = importlib.util.spec_from_loader(loader.name, loader)
                    module = importlib.util.module_from_spec(spec)
                    loader.exec_module(module)
                    module = getattr(module, module_fullname.split(".")[-1])
                string_to_render = module.TEMPLATE_STR

        if self._pyHtmlGui.auto_reload is True:
            self._add_file_to_monitor(file_to_monitor, item.__class__.__name__)

        # replace pyhtmlgui.call(python_function, "arg1") with pyhtmlgui.call({{_create_py_function_reference(python_function)}}, "arg1")
        # this a a convinience function is user does not have to type the annoying stuff and functions look cleaner
        parts = re.split(r'(>| |\(|=|\"|\'|\n|\r|\t|;)(pyhtmlgui\.call\(.+?)(,|\))', string_to_render)
        for i in [i for i, part in enumerate(parts) if part.startswith("pyhtmlgui.call(")]:
            parts[i] = "pyhtmlgui.call({{_create_py_function_reference(%s)}}" % parts[i][15:]
        string_to_render = "".join(parts)

        # use \pyhtmlgui.call to excape pyhtmlgui.call in case of for example <div>Usage: pyhtmlgui.call(this.foobar, "arg1") </div> where the function should not be called
        string_to_render = string_to_render.replace('\pyhtmlgui.call(', 'pyhtmlgui.call(')
        try:
            self._templateCache[item.__class__.__name__] = self._templateEnv.from_string(string_to_render)
        except Exception as e:
            msg = "Failed to load Template "
            if item.TEMPLATE_FILE is not None:
                msg += "from File '%s': " % item.TEMPLATE_FILE
            else:
                msg += "from Class '%s': " % item.__class__.__name__
            msg += " %s" % e
            raise Exception(msg)

        return self._templateCache[item.__class__.__name__]

    def clear_template_cache(self, classnames=None):
        if classnames is None:
            self._templateCache = {}
        else:
            for classname in classnames:
                try:
                    del self._templateCache[classname]
                except:
                    pass

    def _add_child(self, child):
        self._children.add(child)

    def _remove_child(self, child):
        self._children.remove(child)


class JavascriptCallResult():
    def __init__(self, call_id):
        self.call_id = call_id
        self.websocketInstances = weakref.WeakSet()
        self._results_missing = 0
        self._callback = None

    def add_call(self, websocketInstance):
        websocketInstance.javascript_call_result_queues[self.call_id] = queue.Queue()
        self.websocketInstances.add(websocketInstance)
        self._results_missing += 1

    def result_received(self, websocketInstance, result):
        websocketInstance.javascript_call_result_queues[self.call_id].put(result)
        self._results_missing -= 1
        if self._results_missing == 0 and self._callback is not None:
            self._callback(self._collect_results())

    def _collect_results(self):
        results = []
        for websocketInstance in self.websocketInstances:
            results.append(websocketInstance.javascript_call_result_queues[self.call_id].get())
            del websocketInstance.javascript_call_result_queues[self.call_id]
            del websocketInstance.javascript_call_result_objects[
                self.call_id]  # remove ourself from websocket callback list
        errors = [result["error"] for result in results if "error" in result]
        if len(errors) > 0:
            msg = "%s of %s connected frontends returned an error\n" % (len(errors), len(results))
            msg += "\n".join(errors)
            raise Exception(msg)
        return [result["value"] for result in results]

    def __call__(self, callback=None):
        if callback is None:
            return self._collect_results()
        else:
            if self._results_missing == 0:
                callback(self._collect_results())
            else:
                self._callback = callback



from .observable import Observable


class ObservableList(list, Observable):

    def __init__(self, *args, **kwargs):
        Observable.__init__(self)
        list.__init__(self, *args, **kwargs)

    def __iadd__(self, other):
        self.extend(other)
        return self

    def append(self, value):
        index = len(self)
        list.append(self, value)
        self.notify_observers(action="append", index=index, item=value)

    def insert(self, index, value):
        list.insert(self, index, value)
        if index > len(self) - 1:
            index = len(self) - 1
        self.notify_observers(action="insert", index=index, item=value)

    def __setitem__(self, key, value):
        if (type(key) is slice):
            index = key.start
        else:
            index = key
        oldItem = list.__getitem__(self, key)
        list.__setitem__(self, key, value)
        self.notify_observers(action="setitem", index=index, oldItem=oldItem, newItem=value)

    def __delitem__(self, i):
        if isinstance(i, slice):
            index = i.start
        else:
            index = i
        item = list.__getitem__(self, i)
        list.__delitem__(self, i)
        self.notify_observers(action="delitem", index=index, item=item)

    def extend(self, seq):
        insertIndex = len(self)
        list.extend(self, seq)
        self.notify_observers(action="extend", index=insertIndex, items=seq)

    def pop(self, index=-1):
        removedIndex = index
        if (index == -1):
            removedIndex = len(self) - 1
        value = list.pop(self, index)
        self.notify_observers(action="pop", index=removedIndex, item=value)
        return value

    def remove(self, obj):
        index = self.index(obj)
        list.remove(self, obj)
        self.notify_observers(action="remove", index=index, item=obj)

    def sort(self, **kwargs):
        list.sort(self, **kwargs)
        self.notify_observers(action="sort")

    def reverse(self):
        list.reverse(self)
        self.notify_observers(action="reverse")



import weakref

class EventSet():
    def __init__(self):
        self.events = []

    def add(self, observableObject, targetFunction):
        try:
            if not callable(observableObject.attachObserver) or not callable(observableObject.detachObserver):
                raise Exception("object type '%s' can not be observed" % type(observableObject))
        except:
            raise  Exception("object type '%s' can not be observed" % type(observableObject))

        if type(observableObject) != weakref.ref:
            observableObject = weakref.ref(observableObject)
        if type(targetFunction) != weakref.WeakMethod:
            targetFunction = weakref.WeakMethod(targetFunction)
        self.events.append([observableObject, targetFunction])

    def remove(self, observableObject, targetFunction ):
        to_remove = []
        for e in self.events:
            e_observableObject, e_targetFunction = e
            if observableObject == e_observableObject():
                if targetFunction is None or targetFunction == e_targetFunction():
                    to_remove.append(e)
        for e in to_remove:
            self.events.remove(e)

    def attach_all(self):
        for observable, target in self.events:
            try:
                observable().attachObserver(target())  # resolve weak references
            except:
                pass

    def detach_all(self):
        for observable, target in self.events:
            try:
                observable().detachObserver(target())  # resolve weak references
            except:
                pass


import json
import logging
import os
import shutil
import sys
import threading
import time
import traceback
import uuid

import bottle
import bottle_websocket
import jinja2
from gevent import pywsgi
from geventwebsocket.handler import WebSocketHandler
from geventwebsocket.logging import create_logger

from .lib import Browser
from .pyhtmlguiInstance import PyHtmlGuiInstance


class PyHtmlGui():
    def __init__(self,
                 appInstance,  # Some object (eg. main program class instance), passed to appViewClass as obj on launch
                 appViewClass,  # A class that Inherits from PyHtmlView
                 static_dir="",  # static files, css, img go here
                 template_dir="",  # main.html and other html goes here
                 electron_app_dir=None,
                 # in case we use electron, this is the electron.js file we launch, default file is in pyHtmlGui/assets/electron/main.py
                 base_template="pyHtmlGuiBase.html",
                 # pyHtmlGuiBase in pyHtmlGui/assets/templates, or custom file in app templates dir
                 on_frontend_ready=None,  # If gui connects call this
                 on_frontend_exit=None,  # If gui disconnects call this
                 size=(800, 600),  # window size
                 position=None,  # window position
                 mode="chrome",  # chrome | electron
                 executable=None,  # path to chrome/electron executable, if needed
                 listen_host="127.0.0.1",
                 listen_port=8000,
                 shared_secret="",  # use "" to automatically generate a uid internally, use None to disable token
                 auto_reload=False,  # for development, monitor files and reload while app is running
                 single_instance=True # create only one instance and share it between all connected websockets, this is the default, so there is only one instance of appViewClass shared by all connected frntends
                 ):
        self.appViewClass = appViewClass
        self.appInstance = appInstance
        self.static_dir = os.path.abspath(static_dir)
        self.template_dir = os.path.abspath(template_dir)
        self.on_frontend_ready_callback = on_frontend_ready
        self.on_frontend_exit_callback = on_frontend_exit
        self.base_template = base_template
        self.size = size
        self.position = position
        self.mode = mode
        self.executable = executable
        self.electron_app_dir = electron_app_dir
        if shared_secret == "":
            self.shared_secret = "%s" % uuid.uuid4()
        elif shared_secret is None:
            self.shared_secret = None
        else:
            self.shared_secret = shared_secret
        self.listen_host = listen_host
        self.listen_port = listen_port
        self.auto_reload = auto_reload
        self.single_instance = single_instance

        if getattr(sys, 'frozen', False) is True:  # check if we are bundled by pyinstaller
            # noinspection PyProtectedMember
            lib_dir = os.path.join(sys._MEIPASS, "pyhtmlgui")
            self.auto_reload = False
        else:
            lib_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)))

        self._template_dir = os.path.join(lib_dir, "assets", "templates")
        self._electron_dir = os.path.join(lib_dir, "assets", "electron")
        if self.electron_app_dir is None:  # use default internal app
            self.electron_app_dir = self._electron_dir
        else:
            if getattr(sys, 'frozen',
                       False) is not True:  # if we are not frozen, copy our internal lib to the electron target dir if it does not exist
                for f in ["pyhtmlgui.js", "main.js"]:
                    lib_js_target = os.path.join(self.electron_app_dir, f)
                    lib_js_source = os.path.join(self._electron_dir, f)
                    if not os.path.exists(lib_js_target):
                        shutil.copyfile(lib_js_source, lib_js_target)

        if not os.path.exists(self.template_dir):
            raise Exception("Template dir '%s' not found" % self.template_dir)
        if not os.path.exists(self.static_dir):
            raise Exception("Static dir '%s' not found" % self.static_dir)

        self._templateLoader = jinja2.FileSystemLoader(searchpath=[self._template_dir, self.template_dir])
        self._templateEnv = jinja2.Environment(loader=self._templateLoader)

        self._gui_instances = []

        self._token_cookie = "%s" % uuid.uuid4()
        self._token_csrf = "%s" % uuid.uuid4()

        bottle.route("/")(self._main_html)
        bottle.route("/static/<path:path>")(self._static)
        bottle.route("/ws", apply=[bottle_websocket.websocket])(self._websocket)
        self._browser = None
        self._server = MyWebSocketServer(host=self.listen_host, port=self.listen_port)

        self._file_monitoring = {}
        if self.auto_reload is True:
            t = threading.Thread(target=self._monitoring_thread, daemon=True)
            t.start()

    def start(self, show_frontend=False, block=True):
        if show_frontend is True:
            self.show()

        if block is True:
            self._server.run(bottle.default_app())
        else:
            t = threading.Thread(target=self._server.run, args=[bottle.default_app()], daemon=True)
            t.start()

    def stop(self):
        try:
            self._server.server.stop()
        except:
            pass

    def show(self):
        env = None
        if self.mode == "electron":
            args = [self.electron_app_dir, ]
            env = os.environ.copy()
            env.update({
                "PYHTMLGUI_HOST": self.listen_host,
                "PYHTMLGUI_PORT": "%s" % self.listen_port,
                "PYHTMLGUI_SECRET": self.shared_secret,
            })
        else:
            args = ["%s:%s" % (self.listen_host, self.listen_port), ]
            if self.shared_secret is not None:
                args[0] = "%s?token=%s" % (args[0], self.shared_secret)
        if self._browser is None:
            self._browser = Browser(self.mode, self.executable)
        self._browser.open(args, env=env)

    # /main.html
    def _main_html(self):
        if self.shared_secret is not None and bottle.request.query.token != self.shared_secret:
            return bottle.HTTPResponse(status=403)
        template = self._templateEnv.get_template(self.base_template)
        response = bottle.HTTPResponse(template.render({
            "csrf_token": self._token_csrf,
            "start_size": json.dumps(self.size),
            "start_position": json.dumps(self.position),
        }))
        response.set_header('Cache-Control', 'no-store')
        response.set_header('Set-Cookie', 'token=%s' % self._token_cookie)
        return response

    # /static/<path>/<path>
    def _static(self, path):
        if bottle.request.get_cookie("token") != self._token_cookie:
            return bottle.HTTPResponse(status=403)
        response = bottle.static_file(path, root=self.static_dir)
        response.set_header("Cache-Control", "public, max-age=36000")
        return response

    # /ws
    def _websocket(self, ws):
        if bottle.request.get_cookie("token") != self._token_cookie:
            return bottle.HTTPResponse(status=403)
        if bottle.request.query.token != self._token_csrf:
            return bottle.HTTPResponse(status=403)
        websocketInstance = WebsocketInstance(ws)
        instance = self._get_gui_instance()
        instance.websocket_loop(websocketInstance)  # loop while connected
        self._release_gui_instance(instance)

    def _on_frontend_ready(self, pyHtmlGuiInstance):  # called by pyHtmlGuiInstance on frontend ready
        if self.on_frontend_ready_callback is not None:
            nr_of_active_frontends = sum([instance.instances_count for instance in self._gui_instances])
            self.on_frontend_ready_callback(pyHtmlGuiInstance, nr_of_active_frontends)

    def _get_gui_instance(self):
        if self.single_instance is True:
            if len(self._gui_instances) == 0:
                self._gui_instances.append(PyHtmlGuiInstance(self))
            instance = self._gui_instances[0]
        else:
            instance = PyHtmlGuiInstance(self)
            self._gui_instances.append(instance)
        return instance

    def _release_gui_instance(self, instance):
        if instance.instances_count == 0:
            self._gui_instances.remove(instance)

        if self.on_frontend_exit_callback is not None:
            self.on_frontend_exit_callback(instance,
                                           sum([instance.instances_count for instance in self._gui_instances]))

    def _add_file_to_monitor(self, file_to_monitor, class_name):
        if self.auto_reload is False:
            return
        if file_to_monitor not in self._file_monitoring:
            last_changed = os.path.getmtime(file_to_monitor)
            self._file_monitoring[file_to_monitor] = {
                "file_to_monitor": file_to_monitor,
                "last_changed": last_changed,
                "class_names": set(),
            }
        self._file_monitoring[file_to_monitor]["class_names"].add(class_name)

    def _monitoring_thread(self):
        while self.auto_reload is True:
            time.sleep(5)
            has_changed = False
            classed_to_reload = []
            for file_to_monitor in self._file_monitoring:
                data = self._file_monitoring[file_to_monitor]
                current_ts = os.path.getmtime(data["file_to_monitor"])
                if current_ts != data["last_changed"]:
                    self._file_monitoring[file_to_monitor]["last_changed"] = current_ts
                    classed_to_reload.extend(self._file_monitoring[file_to_monitor]["class_names"])
                    has_changed = True
            if has_changed is True:
                for instance in self._gui_instances:
                    instance.clear_template_cache(classed_to_reload)
                    try:
                        instance.update()
                    except Exception as e:
                        print("Failed to Update")
                        print(traceback.format_exc())


class MyWebSocketServer(bottle.ServerAdapter):
    def run(self, handler):
        self.server = pywsgi.WSGIServer((self.host, self.port), handler, handler_class=WebSocketHandler)
        if not self.quiet:
            self.server.logger = create_logger('geventwebsocket.logging')
            self.server.logger.setLevel(logging.INFO)
            self.server.logger.addHandler(logging.StreamHandler())
        self.server.serve_forever()


class WebsocketInstance():
    def __init__(self, ws):
        self.ws = ws
        self.javascript_call_result_queues = {}
        self.javascript_call_result_objects = {}



import types
import weakref
import uuid
import traceback
from threading import Lock
from .lib import EventSet
import time

class PyHtmlView():
    TEMPLATE_FILE = None
    TEMPLATE_STR = None
    WRAPPER_ELEMENT = "pyHtmlView"

    def __init__(self, observedObject, parentView):

        # Component state and data
        self.uid = "%s" % uuid.uuid4()
        self.is_visible = False
        self._children = weakref.WeakSet()

        # Weak references to  observedObject, parentView
        if type(parentView) == weakref.ref:
            parentView = parentView()
        if type(observedObject) == weakref.ref:
            observedObject = observedObject()

        self._observedObject_wref = None
        self._parentView_wref = None
        self._observedObject = None # this is replace for a short time on render by the actual resolved object
        if observedObject is not None:
            self._observedObject_wref = weakref.ref(observedObject,self._on_observedObject_died)  # non gui object we reprecent or talk to
        if parentView is not None:
            parentView._add_child(self)
            self._parentView_wref = weakref.ref(parentView, self._on_parentView_died)  # parent of this element

        # get template loader function from parent
        self._get_template  = parentView._get_template
        self.__was_rendered__ = True
        self.__last_rendered = None # timestamp of last rendering, for debug only

        self._call_javascript = parentView._call_javascript # root element gets this supplied by pyhtmlgui lib, else none
        self.events = EventSet()

        #attach default observation event and detach because default componens is invisible until rendered
        if self._on_observedObject_updated is not None:
            try:
                self.events.add(self.observedObject, self._on_observedObject_updated)
            except Exception as e: # detach all will thow an exception if the event can not be attached
                print(e)
                print("object type '%s' can not be observed" % type(observedObject))



    @property
    def observedObject(self):
        if self._observedObject is not None:
            return self._observedObject
        return self._observedObject_wref()

    @property
    def parentView(self):
        return self._parentView_wref()

    def _on_observedObject_updated(self, source, **kwargs):
        self.update()

    def _on_observedObject_died(self, wr):
        print("observedObject died", wr)
        #raise NotImplementedError()

    def _on_parentView_died(self, wr):
        print("Parent died", wr)
        raise NotImplementedError()

    # return html string rendered from template
    # automatically set component to visible
    def render(self):
        html = self._inner_html()
        if html == None:
            return None
        self.__last_rendered = time.time()
        if self.WRAPPER_ELEMENT is None:
            return html
        else:
            return "<%s id='%s' data-pyhtmlgui-class='%s'>%s</%s>" % (self.WRAPPER_ELEMENT, self.uid, self.__class__.__name__, html, self.WRAPPER_ELEMENT)

    def _inner_html(self):
        self._observedObject = self.observedObject # receive hard reference to obj to it does not die on us while rendering
        if self._observedObject is None:
            print("Observed oject died before render")
            return None
        self.set_visible(True) # It should be ok to set visible here, althou this is befor the rendered object actually appeart in the dom. However, it should arrive at to dom before any other things that might be triggered because the object is visible, because of the websocket event loop
        for child in self._children: child.__was_rendered__ = False
        try:
            html = self._get_template(self).render({"this": self})
        except Exception as e:
            tb = traceback.format_exc()
            msg = " Exception while rendering Template: %s\n" % self.__class__.__name__
            msg += " %s" % tb.replace("\n", "\n  ").strip()
            self.call_javascript("pyhtmlgui.debug_msg", [msg])
            html = msg
            print(msg)

        [c.set_visible(False) for c in self._children if c.__was_rendered__ is False and c.is_visible is True] # set children that have not been rendered in last pass to invisible
        self.__was_rendered__ = True
        self._observedObject = None # remove hard reference to observedobject, it may die now
        return html

    # update rendered component in place, must be visibie
    def update(self):
        if self.is_visible is True:
            html_content = self.render()
            if html_content is not None: # object might have died, in that case don't render
                self.call_javascript("pyhtmlgui.replace_element", [self.uid, html_content], skip_results=True)
        else:
            raise Exception("Can't update invisible components")

    # detach events, remove from parent, remove from frontend if is visible
    def delete(self, already_removed_from_dom = False):
        self.events.detach_all()
        if self.is_visible is True and already_removed_from_dom is False:
            self.call_javascript("pyhtmlgui.remove_element", [self.uid], skip_results=True)
        self.parentView._remove_child(self)
        for child in self._children:
            child.delete(already_removed_from_dom = True)

    # insert this rendered element into parent node at index
    def insert_element(self, index):
        html_content = self.render()
        if html_content is not None:  # object might have died, in that case don't render
            self.call_javascript("pyhtmlgui.insert_element", [self.parentView.uid, index, html_content], skip_results=True)

    # function so we have function arguments name completion in editor, in theorie we could directly use self._call_javascript
    def call_javascript(self, js_function_name, args, skip_results=False):
        return self._call_javascript(js_function_name, args, skip_results)

    # this is a convinience function, you could also call the subcall directly
    def eval_javascript(self, script, skip_results = False, **kwargs):
        if self.is_visible is False:
            raise Exception("Can't javascript_call invisible components")
        return self.call_javascript("pyhtmlgui.eval_script", [script, kwargs], skip_results=skip_results)

    def eval_javascript_electron(self, script, skip_results = False, **kwargs):
        if self.is_visible is False:
            raise Exception("Can't javascript_call invisible components")
        return self.call_javascript("electron.eval_script", [script, kwargs], skip_results=skip_results)

    # set component and childens visibility
    # components that are not visible get their events detached
    def set_visible(self, visible):
        if self.is_visible == visible:
            return

        if visible is False:
            self.is_visible = False
            self.events.detach_all()
            for child in self._children:
                child.set_visible(False)
        else:
            self.is_visible = True
            self.events.attach_all()# don't set children to visible here, because that happens if their render() function is called

    def _add_child(self, child):
        self._children.add(child)

    def _remove_child(self, child):
        self._children.remove(child)



from .weakFunctionReferences import WeakFunctionReferences


class Observable():
    def __init__(self):
        self._observers = WeakFunctionReferences()

    def attach_observer(self, target_function):
        self._observers.add(target_function)

    def detach_observer(self, target_function):
        self._observers.remove(target_function)

    def notify_observers(self, **kwargs):
        for target_function in self._observers.get_all():
            target_function(self, **kwargs)



import types
from threading import Lock
from .pyhtmlview import PyHtmlView


class ObservableListView(PyHtmlView):
    TEMPLATE_STR = '''
        {% for item in this.get_items() %}
            {{ item.render()}}
        {% endfor %}
    '''

    def __init__(self, observedObject, parentView, item_class, wrapper_element=PyHtmlView.WRAPPER_ELEMENT,
                 sort_lambda=None, sort_reverse=False, filter=None, **kwargs):
        self._item_class = item_class
        self.WRAPPER_ELEMENT = wrapper_element
        self._kwargs = kwargs
        self._wrapped_data = []
        self._wrapped_data_lock = Lock()
        self.sort_lambda = sort_lambda
        self.sort_reverse = sort_reverse
        self.filter = filter
        if self.filter is None:
            self.filter = lambda x: False
        super().__init__(observedObject, parentView)

    def get_items(self):
        data = [w for w in self._wrapped_data if self.filter(w) is False]
        if self.sort_lambda is None:
            return data
        else:
            return sorted(data, key=self.sort_lambda, reverse=self.sort_reverse)

    def set_visible(self, visible):
        if self.is_visible == visible:  # not changed
            return
        self._wrapped_data_lock.acquire()
        super().set_visible(visible)
        self._wrapped_data = []
        if self.is_visible is True:  # was set to invisible
            for item in self.observedObject:
                self._wrapped_data.append(self._create_item(item))
        self._wrapped_data_lock.release()

    def _create_item(self, item):
        obj = self._item_class(item, self, **self._kwargs)
        obj.loop_index = types.MethodType(lambda x: x.parentView._loop_index(x), obj)
        obj.loop_index_used = False
        return obj

    def _on_observedObject_updated(self, source, **kwargs):
        try:
            self._wrapped_data_lock.acquire()

            if kwargs["action"] in ["append", "insert"]:
                obj = self._create_item(kwargs["item"])
                self._wrapped_data.insert(kwargs["index"], obj)
                if self.filter(obj) is False:
                    if self.insert_element(kwargs["index"], obj) is False:
                        self._wrapped_data.remove(obj)
                    else:
                        [item.update() for item in self._wrapped_data[kwargs["index"] + 1:] if
                         item.loop_index_used == True]  # update items that use the loop index

            if kwargs["action"] == "setitem":
                self._wrapped_data[kwargs["index"]].delete()  # unrender
                obj = self._create_item(kwargs["newItem"])
                self._wrapped_data[kwargs["index"]] = obj
                if self.filter(obj) is False:
                    if self.insert_element(kwargs["index"], obj) is False:
                        self._wrapped_data.remove(obj)

            if kwargs["action"] == "extend":
                current_index = kwargs["index"]
                for item in kwargs["items"]:
                    obj = self._create_item(item)
                    self._wrapped_data.insert(current_index, obj)
                    if self.filter(obj) is False:
                        if self.insert_element(kwargs["index"], obj) is False:
                            self._wrapped_data.remove(obj)
                        else:
                            current_index += 1
                [item.update() for item in self._wrapped_data[current_index:] if item.loop_index_used == True]

            if kwargs["action"] in ["remove", "pop", "delitem"]:
                self._wrapped_data[kwargs["index"]].delete()
                del self._wrapped_data[kwargs["index"]]
                [item.update() for item in self._wrapped_data[kwargs["index"]:] if item.loop_index_used == True]

            if kwargs["action"] == "sort":
                self.update()
            if kwargs["action"] == "reverse":
                self.update()
        finally:
            self._wrapped_data_lock.release()

    def _loop_index(self, element):
        element.loop_index_used = True
        return self._wrapped_data.index(element)
